from dataclasses import dataclass
import pickle
from typing import Any, Dict, Optional
from .active_record import activerecord

from .database import (
    Database,
    Mapping,
    from_data_class,
    updates,
    fields,
    placeholders,
)


@activerecord
@dataclass
class SagaDTO:
    id: int
    name: str
    instance: str
    step: str
    input: Any
    state: Any
    runner: str
    delay: float
    status: int
    error: Exception


class SagaDAL:
    class ConcurrencyException(Exception):
        pass

    def __init__(self, db: Database):
        self._db = db

    async def new_saga(self, saga: SagaDTO):
        """Return a new saga

        Examples:
        >>> dal = getfixture('dal')
        >>> saga = run_async(
        ...   dal.new_saga(
        ...     SagaDTO(
        ...       id=1,
        ...       name="test",
        ...       instance="random1",
        ...       step="step",
        ...       input="abc",
        ...       state="abc",
        ...       runner="runner",
        ...       delay=None,
        ...       status=1,
        ...       error=None,
        ...     )
        ...   )
        ... )
        >>> saga.name
        'test'
        """

        dao = from_data_class(
            saga,
            input=lambda x: pickle.dumps(x),
            state=lambda x: pickle.dumps(x),
            instance_id=("instance", lambda x: x),
            runner_id=("runner", lambda x: x),
        )
        delay = dao.pop("delay", 0)
        dao.pop("id", 0)

        id = await self._db.fetchval(
            f"INSERT INTO saga.instance ({fields(dao)},start_after) VALUES({placeholders(dao)},now()::timestamp+interval '1 seconds'* :delay) RETURNING id",
            **dao,
            delay=delay or 0,
        )
        saga.id = id
        saga.clear()
        return saga

    async def get_instance(self, name, instance) -> Optional[Dict]:
        """Return id and state for a pending saga with given name and instance.

        Examples
        >>> dal = getfixture('dal')
        >>> run_async(dal.get_instance('test', 'random'))
        {'id': 1, 'state': 'abc'}
        """
        return await self._db.fetchrow(
            "SELECT id, state FROM saga.instance WHERE name = :name AND instance_id = :instance AND status = 0",
            Mapping(dict, state=lambda s: pickle.loads(s)),
            name=name,
            instance=instance,
        )

    async def get_state(self, name, instance):
        """Return state for a saga with given name and instance.

        Examples
        >>> dal = getfixture('dal')
        >>> run_async(dal.get_state('test', 'random'))
        'abc'
        """
        ret = await self._db.fetchval(
            "SELECT state FROM saga.instance WHERE name = :name AND instance_id = :instance",
            name=name,
            instance=instance,
        )
        return pickle.loads(ret)

    async def update_processing(self, saga: SagaDTO):
        """Update the saga if it is in processing status and runner matches the value in *saga*.
        Otherwise it will throw ConcurrencyException.

        Examples
        >>> dal = getfixture('dal')
        >>> saga = SagaDTO(id=3,name="test",instance="random3",step="step",input="abc",state="abc",runner='runner',delay=None,status=1,error=None,)
        >>> saga = run_async(dal.new_saga(saga))
        >>> saga.input="def"
        >>> saga.state="def"
        >>> run_async(dal.update_processing(saga))
        >>> run_async(dal.get_state('test', 'random3'))
        'def'
        >>> saga.state = '111'
        >>> saga.runner='runner123'
        >>> with raises(SagaDAL.ConcurrencyException): run_async(dal.update_processing(saga))
        """
        dao = from_data_class(
            saga,
            saga.changes(),
            input=lambda x: pickle.dumps(x),
            state=lambda x: pickle.dumps(x),
            instance_id=("instance", lambda x: x),
        )
        saga.clear()

        # runner cannot be updated
        dao.pop("runner", None)
        # delay will be covered manually
        dao.pop("delay", 0)

        if len(dao) == 0:
            return None

        id = await self._db.fetchval(
            f"UPDATE saga.instance SET {updates(dao)},start_after=now()::timestamp+interval '1 seconds'* :delay WHERE id = :id AND status = 1 AND runner_id = :runner_id RETURNING id",
            **dao,
            delay=saga.delay or 0,
            id=saga.id,
            runner_id=saga.runner,
        )
        if id == None:
            raise SagaDAL.ConcurrencyException()

    async def update_pending(self, saga: SagaDTO):
        """Return state for a saga with given name and instance.

        Examples
        >>> dal = getfixture('dal')
        >>> saga = SagaDTO(id=4,name="test",instance="random4",step="step",input="abc",state="abc",runner='runner',delay=None,status=0,error=None,)
        >>> saga = run_async(dal.new_saga(saga))
        >>> saga.runner="runner4"
        >>> saga.state="def"
        >>> run_async(dal.update_pending(saga))
        >>> run_async(dal.get_state('test', 'random4'))
        'def'

        If the saga isn't in pending status, it will throw a ConcurrencyException

        >>> saga.status = 1
        >>> saga.runner = 'my_runner'
        >>> run_async(dal.update_pending(saga))
        >>> saga.state = 'aaa'
        >>> with raises(SagaDAL.ConcurrencyException): run_async(dal.update_pending(saga))
        """
        if len(saga.changes()) == 0:
            return None

        dao = from_data_class(
            saga,
            saga.changes(),
            input=lambda x: pickle.dumps(x),
            state=lambda x: pickle.dumps(x),
            instance_id=("instance", lambda x: x),
            runner_id=("runner", lambda x: x),
        )
        saga.clear()

        # id and delay will be covered manually
        dao.pop("id", 0)
        dao.pop("delay", 0)

        id = await self._db.fetchval(
            f"UPDATE saga.instance SET {updates(dao)},start_after=now()::timestamp+interval '1 seconds'* :delay WHERE id = :id AND status = 0 RETURNING id",
            **dao,
            delay=saga.delay or 0,
            id=saga.id,
        )
        if id == None:
            raise SagaDAL.ConcurrencyException()

    async def get_pending_saga(self, name, runner, batch_size):
        """Return pending sagas that are scheduled to be started, and set them to be in_progress

        *name*: saga name
        *runner*: runner_id to be set to for the pending sagas
        *batch_size*: number of pending sagas to retrieve

        Examples:
        >>> dal = getfixture('dal')
        >>> run_async(dal.get_pending_saga("random_name", "runner1", 100))
        []
        """
        return await self._db.fetch(
            """
WITH cte AS (
  SELECT id FROM saga.instance
  WHERE name = :name AND status = 0 AND start_after < now()::timestamp
  ORDER BY start_after
  LIMIT :limit
)
UPDATE saga.instance i
SET    status = 1, runner_id = :runner
FROM   cte
WHERE  cte.id = i.id AND i.status = 0
RETURNING i.*
""",
            Mapping(
                SagaDTO,
                input=lambda x: pickle.loads(x),
                state=lambda x: pickle.loads(x),
                instance=("instance_id", lambda x: x),
                runner=("runner_id", lambda x: x),
                delay=("start_after", lambda x: 0),
            ),
            name=name,
            limit=batch_size,
            runner=runner,
        )

    async def keep_alive(self, runner, timeout):
        """Keep current runner alive and clean up timed out runners based on *timeout* value

        Examples:
        >>> dal = getfixture('dal')
        >>> run_async(dal.keep_alive("my_runner", 5))
        """
        await self._db.execute(
            "call saga.heart_beat(:runner, :timeout)",
            runner=runner,
            timeout=timeout,
        )

    async def init_db(self):
        """Initialize the database"""
        await self._db.execute(
            """
BEGIN;
CREATE SCHEMA IF NOT EXISTS saga;
COMMIT;

BEGIN;
CREATE TABLE IF NOT EXISTS saga.instance (
  id BIGSERIAL PRIMARY KEY,
  name VARCHAR(100)  NOT NULL,
  instance_id VARCHAR(100) NOT NULL,
  step VARCHAR(100) NOT NULL,
  input BYTEA,
  state BYTEA,
  runner_id VARCHAR(100),
  start_after TIMESTAMP WITHOUT TIME ZONE,
  status INT NOT NULL, -- 0: pending, 1: in_progress, 2: complete, 3: error
  error TEXT
);
CREATE TABLE IF NOT EXISTS saga.runner (
  id BIGSERIAL PRIMARY key,
  runner_id VARCHAR(100),
  last_seen TIMESTAMP WITHOUT TIME ZONE
);
COMMIT;

BEGIN;
CREATE UNIQUE INDEX IF NOT EXISTS IX_instance_name_instance_id ON saga.instance(name, instance_id); -- ok
CREATE INDEX IF NOT EXISTS IX_instance_runner_id_status ON saga.instance(runner_id, status);
CREATE INDEX IF NOT EXISTS IX_saga_runner_runner_id ON saga.runner(runner_id);
CREATE INDEX IF NOT EXISTS IX_saga_runner_last_seen ON saga.runner(last_seen);

DROP PROCEDURE IF EXISTS saga.clean_up_dead_runner;
CREATE OR REPLACE PROCEDURE saga.heart_beat(current_runner VARCHAR(100), timeout int)
AS $$
  DECLARE 
    r_id BIGINT;
    now TIMESTAMP WITHOUT TIME ZONE = NOW();
  BEGIN
    UPDATE saga.runner SET last_seen = now WHERE runner_id = current_runner RETURNING id INTO r_id;
    IF r_id IS NULL THEN
      INSERT INTO saga.runner (runner_id, last_seen) VALUES(current_runner, now);
    END IF;
    CALL saga.clean_up_dead_runner(timeout);
  END
$$
LANGUAGE plpgsql;

DROP PROCEDURE IF EXISTS saga.clean_up_dead_runner;
CREATE OR REPLACE PROCEDURE saga.clean_up_dead_runner(timeout int)
AS $$
  DECLARE 
    now TIMESTAMP WITHOUT TIME ZONE = NOW();
  BEGIN
    DELETE FROM saga.runner WHERE last_seen < now - timeout * interval '1 seconds';
    UPDATE saga.instance AS s
    SET status = 0, start_after = now
    FROM saga.instance AS i
    LEFT OUTER JOIN saga.runner AS c
    ON i.runner_id = c.runner_id
    WHERE c.id IS NULL AND s.id = i.id AND s.status = 1;
  END
$$
LANGUAGE plpgsql;
COMMIT;
        """
        )



from asyncio import (
    CancelledError,
    Event,
    gather,
    sleep,
)
from dataclasses import dataclass
from time import time
from traceback import format_exc
from typing import Any, AsyncGenerator, Callable, Coroutine, Optional, TypeVar, Union
import uuid
from .dal import SagaDAL, SagaDTO
from .logger import LogEntry, Logger
from .stream_utils import AsyncIterator, merge, start_loop, poll
from .database import Database


TInput = TypeVar("TInput")
TState = TypeVar("TState")


@dataclass
class SagaReturn:
    step: str
    input: TInput
    state: Optional[TState] = None
    delay: float = None


class SagaError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


async def default_log_handler(log: AsyncGenerator[LogEntry, None]):
    async for l in log:
        pass


class Saga:
    def _set_state(dal: SagaDAL, saga: SagaDTO):
        """Return a function to set internal state for a saga"""

        async def func(state):
            saga.state = state
            await dal.update_processing(saga)

        return func

    def _saga_step_decorator_generator(self):
        def saga_step_decorator(
            func: Callable[
                [TInput, TState, Callable[[TState], Coroutine[Any, Any, None]], Logger],
                Union[SagaReturn, Any],
            ]
        ):
            """Decorator for a saga function

            Saga function can either return another saga function for continuation, or any other value as final state to terminate the saga

            Example:
            >>> @saga.step
            ... my_func(input, state, set_state, logger):
            ...   return another_saga_func(input, state, delay=10) # delay execution for 10 seconds

            """

            async def saga_wrapper(saga: SagaDTO):
                try:
                    logger = self._logger.with_context(
                        saga.id, saga.instance, saga.step
                    )
                    logger.log_info("Executing", source="saga_wrapper")
                    ret = await func(
                        saga.input, saga.state, Saga._set_state(self._dal, saga), logger
                    )

                    if isinstance(ret, SagaReturn):
                        step, delay, status = self._get_step_delay_status(
                            ret.step, ret.delay
                        )
                        saga.step = step
                        saga.delay = delay
                        saga.status = status
                        saga.input = ret.input

                        if ret.state != None:
                            saga.state = ret.state
                    else:
                        # if return value isn't a SagaReturn object, it is treated as end of saga
                        saga.status = 2

                        if ret != None:
                            saga.state = ret

                    logger.log_info(
                        "Updating Saga",
                        new_step=saga.step,
                        delay=saga.delay,
                        status=saga.status,
                    )
                except SagaError as e:
                    saga.status = 3
                    saga.error = repr(e)
                    logger.log_info("Returning SagaError", e)
                except Exception as e:
                    # put back to the database for future pickup
                    saga.status = 0
                    logger.log_error(e, format_exc())

                # queue up update saga (to allow retrying)
                self._update_saga_queue.send(
                    {"saga": saga, "time": time(), "retry_count": 0}
                )

            name = func.__name__
            self._function_dict[name] = saga_wrapper

            # Consider the following example:
            #
            # @saga.step
            # def step1(input, state, set_state, logger):
            #    return step2(input)
            #
            # the function will convert the next step into a SagaReturn
            #
            def meta_definition(
                input: TInput, state: Optional[TState] = None, delay: float = None
            ):
                return SagaReturn(name, input, state, delay)

            # set __name__ back to original function name
            meta_definition.__name__ = name
            return meta_definition

        return saga_step_decorator

    def __init__(self, name: str, dal=None, dsn=None):
        self.name = name
        self.runner = uuid.uuid4().hex

        # SagaDAL can be injected in for internal testing. Outside this module, caller will pass in dsn
        self._dal = dal or SagaDAL(Database(dsn))

        self._function_dict = {}
        self._in_mem_sagas = AsyncIterator()
        self._update_saga_queue = AsyncIterator()
        self._event_loop_started = False

        self._logger = Logger(name, self.runner)

        self.step = self._saga_step_decorator_generator()

    async def start(self, instance: str, step, input, state, delay: float = None):
        """Start a new saga

        *instance*: a unique identifier for the saga instance
        *step*: step of the saga. Can be a function or string
        *input*: input to the saga function
        *state*: internal state of saga function
        *delay*: for delay start of a saga function. If not set, the saga will be enqueued for immediate processing (when event loop starts)

        Examples:
        >>> dal = getfixture('dal')
        >>> saga = Saga("deposit", dal)
        >>> @saga.step
        ... def step1(input, state, set_state, logger):
        ...   print(f'step1 started with {input} and {state}')

        >>> run_async(saga.start(uuid.uuid4().hex, step1, 'input', 'state'))
        """
        step, delay, status = self._get_step_delay_status(step, delay)
        saga = SagaDTO(
            0, self.name, instance, step, input, state, self.runner, delay, status, None
        )

        logger = self._logger.with_context(None, instance, step)
        logger.log_info("Starting saga", delay=delay, status=status)
        saga = await self._dal.new_saga(saga)
        self._logger.with_context(saga.id, instance, step).log_info("Started")

        # if status of saga is in_progress (immediate start), send to task queue for consumption
        if saga.status == 1:
            self._in_mem_sagas.send(saga)

    async def get_state(self, instance: str):
        """Return the internal state of a saga

        Examples:
        >>> dal = getfixture('dal')
        >>> saga = Saga("test", dal)
        >>> run_async(saga.get_state('random'))
        'abc'
        """
        self._logger.with_context(None, instance, None).log_trace(
            "Getting state", source="get_state"
        )
        return await self._dal.get_state(self.name, instance)

    async def call(
        self,
        instance: str,
        step: Union[Callable, str],
        input: TInput,
        delay: Optional[float] = None,
    ):
        """Overwrite a saga with new step, input and delay"""

        ret = await self._dal.get_instance(self.name, instance)
        step, delay, status = self._get_step_delay_status(step, delay)

        saga = SagaDTO(
            ret["id"],
            self.name,
            instance,
            None,
            None,
            ret["state"],
            None,
            None,
            None,
            None,
        )

        # SagaDTO is an active record. Setting the values so it can be captured in the changeset.
        saga.step = step
        saga.input = input
        saga.runner = self.runner
        saga.delay = delay
        saga.status = status

        logger = self._logger.with_context(saga.id, instance, saga.step)

        logger.log_trace(
            "Updating", delay=saga.delay, status=saga.status, source="call"
        )

        await self._dal.update_pending(saga)

        # if status of saga is in_progress (immediate start), send to task queue for consumption
        if saga.status == 1:
            self._in_mem_sagas.send(saga)

    def init_db(self):
        return self._dal.init_db()

    async def start_event_loop(
        self,
        batch_size: int,
        low_watermark: int,
        backoff: float,
        available_workers: int,
        available_update_threads: int,
        health_check_timeout: int,
        log_handler: Callable[
            [AsyncGenerator[LogEntry, None]], Coroutine[None, None, None]
        ] = default_log_handler,
    ):
        """Start the event loop to do the following

        1. heart beat to register and keep the instance alive
        2. poll DB for pending sagas. It will *backoff* in seconds if in mem queue size > *low_watermark*, or get less items from db, per *batch_size*
        3. create worker pool to execute saga functions (pool size defined by *available_workers*)
        4. maintain a queue to update saga (and retry). worker threads is defined as *available_update_threads*

        Examples:
        >>> dal = getfixture('dal')
        >>> saga = Saga("event_loop", dal)
        >>> loop = asyncio.get_event_loop().create_task(saga.start_event_loop(10, 10, 0.2, 1, 1, 5))
        >>> @saga.step
        ... async def step123(input, state, set_state, logger):
        ...   loop.cancel()
        >>> run_async(loop)
        """
        cancel = Event()
        self._event_loop_started = True

        async def get_pending_saga():
            try:
                self._logger.log_trace(
                    "Getting pending sagas", source="get_pending_saga"
                )
                return await self._dal.get_pending_saga(
                    self.name, self.runner, batch_size
                )
            except Exception as e:
                self._logger.log_error(e, format_exc(), source="get_pending_saga")
                return []

        async def execute_saga(saga: SagaDTO):
            if not saga.step in self._function_dict:
                e = SagaError("Unrecognized saga step")
                saga.error = repr(e)
                saga.status = 3
                self._logger.log_error(e, source="execute_saga")
                self._update_saga_queue.send((saga, time(), 0))
            else:
                await self._function_dict[saga.step](saga)

        async def update_saga(input):
            saga: SagaDTO = input["saga"]
            retry_count: int = input["retry_count"]

            logger = self._logger.with_context(
                saga.id, saga.instance, saga.step, "update_saga"
            )

            # calibrate delay
            if saga.delay != None and retry_count > 0:
                now = time()
                delay = saga.delay - (now - input["time"])
                saga.delay = 0 if delay < 0 else delay

            try:
                await self._dal.update_processing(saga)
                if saga.status == 1:
                    self._in_mem_sagas.send(saga)
            except SagaDAL.ConcurrencyException:
                logger.log_warn("Update concurrency failure")
                return
            except Exception as e:
                # could be temporary glitch, or db connectivity issues.
                retry_count += 1
                logger.log_error(
                    e, format_exc(), source="update_saga", retry_count=retry_count
                )
                # put back into the queue for retry
                self._update_saga_queue.send(
                    {"saga": saga, "time": now, "retry_count": retry_count}
                )

        # send regular heartbeat to DB
        async def health_check(cancel: Event):
            while not cancel.is_set():
                try:
                    await self._dal.keep_alive(self.runner, health_check_timeout)
                except Exception as e:
                    self._logger.log_error(e, format_exc(), source="health_check")
                    pass
                await sleep(1)

        def end_loop(_):
            self._logger.log_info("Event loop ended")
            self._logger.logstream.end()

        # merge two queues (async generator) - tasks from polling and in-memory ones into a single stream
        task_queue = merge(
            poll(get_pending_saga, low_watermark, batch_size, backoff, cancel),
            self._in_mem_sagas,
        )

        try:
            execute_saga_loop = start_loop(execute_saga, task_queue, available_workers)
            execute_saga_loop.add_done_callback(lambda _: self._update_saga_queue.end())

            update_saga_loop = start_loop(
                update_saga, self._update_saga_queue, available_update_threads
            )

            update_saga_loop.add_done_callback(end_loop)

            await gather(
                health_check(cancel),
                execute_saga_loop,
                update_saga_loop,
                log_handler(self._logger.logstream),
            )
        except CancelledError:
            # asyncio.get_event_loop.stop() will cancel all tasks in event loop, which generates CancelledError
            # seeing this error means event_loop has stopped, so set the flag to break all loops
            cancel.set()
            self._event_loop_started = False
            self._logger.log_info("Event loop cancelled", source="start_event_loop")

    def _get_step_delay_status(self, step, delay: float = None):
        # if delay is not set, set status to be in_progress for immediate consumption
        if self._event_loop_started and delay == None:
            status = 1
        else:
            status = 0

        # convert function into its name
        if callable(step):
            step = step.__name__

        return (step, delay, status)



from collections import deque
from concurrent.futures import FIRST_COMPLETED
from threading import Lock
from asyncio import (
    Future,
    Task,
    Event,
    create_task,
    gather,
    get_event_loop,
    sleep,
    wait,
)
from time import time
from typing import (
    AsyncGenerator,
    Callable,
    Coroutine,
    Generator,
    Generic,
    List,
    TypeVar,
)

T = TypeVar("T")


class AsyncIterator(AsyncGenerator[T, None], Generic[T]):
    """AsyncIterator implements async generator interface for streaming data.

    It also includes a *send* method for caller to push data into the stream.

    Examples:

    >>> ai = AsyncIterator()
    >>> ai.send(123)
    >>> run_async(ai.anext())
    123
    >>> ai.end()
    >>> with raises(StopAsyncIteration): run_async(ai.anext())

    """

    def __init__(self):
        self.__ll = deque()
        # Rule of thumb: never use threading.Lock in async functions
        self.__lock = Lock()
        self.__ll.append(Future())

        async def source(ll: deque):
            while True:
                try:
                    r = await ll[0]
                    ll.popleft()
                    yield r
                except StopAsyncIteration:
                    break

        self.__source = source(self.__ll)

    def send(self, item: T):
        with self.__lock:
            self.__ll.append(Future())
            self.__ll[-2].set_result(item)

    def end(self):
        with self.__lock:
            self.__ll[-1].set_exception(StopAsyncIteration())

    def __aiter__(self):
        return self.__source.__aiter__()

    def __anext__(self):
        with self.__lock:
            return self.__source.__anext__()

    def asend(self, value):
        raise NotImplementedError()

    def athrow(self, **args):
        raise NotImplementedError()

    def anext(self):
        return self.__anext__()


def to_sync(gen: AsyncGenerator[T, None], loop=None):
    """Convert an async generator to a list of futures

    Examples:

    >>> async def gen():
    ...   for i in range(1, 4):
    ...     yield i
    >>> sync = to_sync(gen())
    >>> ret = (run_async(i) for i in sync)
    >>> [next(ret), next(ret), next(ret)]
    [1, 2, 3]
    >>> with raises(StopAsyncIteration): next(ret)
    """

    async def _cont(gen, last_value):
        if last_value != None:
            await last_value
        return await gen.__anext__()

    loop = loop or get_event_loop()
    lock = Lock()
    ret = None
    while True:
        with lock:
            ret = loop.create_task(_cont(gen, ret))
            yield ret


async def _loop_one(
    func: Callable[[T], Coroutine[None, None, None]],
    gen: Generator[Task[T], None, None],
):
    while True:
        item = await next(gen)
        try:
            await func(item)
        except:
            continue


def start_loop(
    func: Callable[[T], Coroutine[None, None, None]],
    gen: AsyncGenerator[T, None],
    size: int,
):
    """Start a worker pool to execute function on items from stream

    *func*: function to process items. Exception will be swallowed.
    *gen*: async stream to read data from
    *size*: pool size
    *cancel*: cancellation event

    Examples:
    >>> async def gen():
    ...   for i in range(1, 1001):
    ...     yield i
    >>> class Data:
    ...   sum_all = 0
    >>> data = Data()
    >>> def worker_generator(data):
    ...   def worker(i):
    ...     data.sum_all += i
    ...   return worker
    >>> _ = run_async(start_loop(worker_generator(data), gen(), 10))
    >>> data.sum_all
    500500
    """
    it = to_sync(gen)
    arr = [_loop_one(func, it) for i in range(0, size)]
    return gather(*arr, return_exceptions=True)


async def poll(
    func: Callable[[], Coroutine[None, None, List[T]]],
    low_watermark: int,
    expected_len: int,
    backoff: float,
    cancel: Event,
):
    """
    Call func periodically to get list of items and return as async stream.

    *func*: async function that returns a list of items
    *low_watermark*: don't poll if remaining items in the queue >= low_watermark
    *expected_len*: back off if *func* returns less item than *expected_len*
    *backoff*: back off time in seconds
    *cancel*: cancellation event

    Examples:
    >>> l = iter([30, 9, 1])
    >>> import time
    >>> async def func():
    ...   now = time.time()
    ...   return [now for i in range(0, next(l))]
    >>> evt = asyncio.Event()
    >>> stream = poll(func, 10, 20, 0.2, evt)
    >>> batch1 = [run_async(stream.__anext__()) for i in range(0, 20)][0]
    >>> batch2 = [run_async(stream.__anext__()) for i in range(0, 10)][-1]
    >>> batch1 == batch2
    True
    >>> batch3 = [run_async(stream.__anext__()) for i in range(0, 10)]
    >>> batch3[0] == batch3[8]
    True
    >>> batch3[9] - batch3[0] > 0.2
    True
    >>> batch3[0] - batch2 < 0.2
    True
    >>> async def cancel_event():
    ...   await asyncio.sleep(0.2)
    ...   evt.set()
    >>> _ = asyncio.get_event_loop().create_task(cancel_event())
    >>> with raises(StopAsyncIteration): run_async(stream.__anext__())
    """

    async def delayed_func(state):
        last_run, last_count = state

        delay = backoff + last_run - time() if last_count < expected_len else -1
        if delay > 0:
            await sleep(delay)

        state[0] = time()
        try:
            ret = await func()
        except:
            ret = []
        state[1] = len(ret)
        return ret

    values: List[T] = []
    task, state = None, [0, expected_len]

    while not cancel.is_set():
        l = len(values)
        if l < low_watermark and task == None:
            task = delayed_func(state)

        if l == 0:
            values, task = await task, None
            continue

        ret, *values = values
        yield ret


async def _coroutine_tuple(task, *args):
    return (await task, *args)


async def merge(*generators, when_exception="IGNORE"):
    """Merge multiple streams into a single stream

    Examples:
    >>> async def gen1():
    ...  for i in range(0, 10):
    ...    await asyncio.sleep(0.01)
    ...    yield i
    >>> async def gen2():
    ...  for i in range(0, 10):
    ...    await asyncio.sleep(0.02)
    ...    yield i
    >>> new_stream = merge(gen1(), gen2())
    >>> sum([run_async(new_stream.__anext__()) for i in range(0, 20)])
    90
    """
    nexts = [create_task(_coroutine_tuple(gen.__anext__(), gen)) for gen in generators]
    while len(nexts):
        done, pending = await wait(nexts, return_when=FIRST_COMPLETED)
        for d in done:
            e = d.exception()
            if (e != None and when_exception == "IGNORE") or isinstance(
                e, StopAsyncIteration
            ):
                continue
            value, gen = d.result()
            pending.add(create_task(_coroutine_tuple(gen.__anext__(), gen)))
            yield value
        nexts = pending



from asyncio import Task, get_event_loop
from dataclasses import asdict
from typing import Any, AsyncContextManager, Dict, List, OrderedDict, overload
import typing
from asyncpg import Connection, Pool, create_pool


class Mapping(typing.Mapping):
    """Maps SQL record to data class

    *cls_type*: the data class type to map record to
    *kw_args*: named argument to define any transformation required before setting the value for data class

    Examples:
    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class MyType:
    ...   a: int
    ...   b: int
    ...   c: int
    >>> # define simple mapping without transformation
    >>> mapping = Mapping(MyType)
    >>> # define mapping with transformation
    >>> mapping = Mapping(MyType, c=lambda x: x + 1)
    >>> # transform database column columnX to c
    >>> mapping = Mapping(MyType, c=('columnX', lambda x: x))
    """

    def __init__(self, cls_type, **kw_args):
        self._cls_type = cls_type
        self._transform = []
        for k in kw_args:
            v = kw_args[k]
            if callable(v):
                new, old, func, remove_flag = k, k, v, False
            else:
                if len(v) == 2:
                    v = (*v, True)

                new, old, func, remove_flag = (k, *v)

            self._transform.append((new, old, func, remove_flag))

    def __iter__(self):
        return iter(self.dict)

    def __getitem__(self, item):
        return self.dict[item]

    def __len__(self):
        return len(self.dict)

    def value(self):
        return self._cls_type(**self.dict)

    def load(self, **kw_args):
        self.dict = dict(kw_args)
        for new, old, func, remove_flag in self._transform:
            self.dict[new] = func(self.dict[old])

            if remove_flag:
                del self.dict[old]

        return self


def fields(m: typing.Mapping) -> str:
    """Return keys in the mapping as a comma separated string

    Example:
    >>> fields(dict(id=1, col1="abc"))
    'id,col1'
    """
    return ",".join([k for k in m])


def placeholders(m: typing.Mapping) -> str:
    """Return keys in the mapping as a comma separated placeholders

    Example:
    >>> placeholders(dict(id=1, col1="abc"))
    ':id,:col1'
    """
    return ",".join([":" + k for k in m])


def updates(m: typing.Mapping, prefix: str = None) -> str:
    """Return keys in *m* in a format acceptable for UPDATE

    Example:
    >>> updates(dict(id=1, col1="abc"), prefix='my_table')
    'my_table.id=:id,my_table.col1=:col1'
    """
    prefix = prefix + "." if prefix != None else ""
    return ",".join([f"{prefix}{k}=:{k}" for k in m])


def from_data_class(value, includes: List[str] = None, **kw_args) -> typing.Mapping:
    """Return a Mapping from *value*, applying functions to the fields

    *includes*: if specified, trim down the returned mapping to only the attributes in *includes*
    Example:
    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class MyType:
    ...   a: int
    ...   b: int
    ...   c: int
    >>> input = MyType(a=1,b=2,c=3)
    >>> from_data_class(input, ['a', 'b'],
    ...   a = lambda x: x+10,
    ...   new_b = ('b', lambda x: x),
    ... )
    OrderedDict([('a', 11), ('new_b', 2)])
    """
    value = include(asdict(value), includes) if includes != None else asdict(value)
    return apply(value, **kw_args)


def apply(value: typing.Mapping, **kw_args) -> typing.Mapping:
    """Return a Mapping, applying functions to the fields

    Keyword argument can be a function or a tuple with the format of (old_field, mapping_function, remove_old_field)
    *remove_old_field* can be omitted (default True)

    Examples:
    >>> apply(dict(id=1, col1="abc", col2='def'),\
            id=lambda x: x+1,\
            col1=lambda x: x+'1',\
            col3=('col2', lambda x: x+'2')\
        )
    {'id': 2, 'col1': 'abc1', 'col3': 'def2'}
    """
    for k in kw_args:
        v = kw_args[k]
        if callable(v):
            new, old, func, remove_flag = k, k, v, False
        else:
            if len(v) == 2:
                v = (*v, True)

            new, old, func, remove_flag = k, *v
        if old in value:
            value[new] = func(value[old])
            if remove_flag:
                del value[old]
    return value


@overload
def include(m: typing.Mapping, arg_list: List[str]) -> OrderedDict:
    ...


def include(m: typing.Mapping, *args) -> OrderedDict:
    """Return a new mapping contains only keys specified in *args*.

    *args* can be strings, comma separated strings or list of strings.

    Examples
    >>> include(dict(id=1,col1="abc",col2="def"), ["col1"], "id")
    OrderedDict([('col1', 'abc'), ('id', 1)])
    """
    new_map = OrderedDict()
    for arg in args:
        if isinstance(arg, str):
            arg = list(map(str.strip, arg.split(",")))

        for k in arg:
            if k in m:
                new_map[k] = m[k]

    return new_map


class _Connection:
    @staticmethod
    def _format_sql(sql: str, **kw_args):
        params_list = []
        for k in kw_args:
            i = len(params_list) + 1
            new_sql = sql.replace(f":{k}", f"${i}")
            if new_sql != sql:
                sql = new_sql
                params_list.append(kw_args[k])
        return (sql, params_list)

    def __init__(self, conn: Connection) -> None:
        self._conn = conn

    async def fetchval(self, sql, col=0, command_timeout: float = None, **kw_args):
        sql, args = self._format_sql(sql, **kw_args)
        return await self._conn.fetchval(
            sql, *args, column=col, timeout=command_timeout
        )

    async def fetch(self, sql, sm: Mapping, command_timeout: float = None, **kw_args):
        sql, args = self._format_sql(sql, **kw_args)
        records = await self._conn.fetch(sql, *args, timeout=command_timeout)
        return [sm.load(**record).value() for record in records]

    async def fetchrow(
        self, sql, sm: Mapping, command_timeout: float = None, **kw_args
    ):
        sql, args = self._format_sql(sql, **kw_args)
        record = await self._conn.fetchrow(sql, *args, timeout=command_timeout)
        if record != None:
            return sm.load(**record).value()
        else:
            return None

    async def execute(self, sql: str, command_timeout: float = None, **kw_args):
        sql, args = self._format_sql(sql, **kw_args)
        return await self._conn.execute(sql, *args, timeout=command_timeout)

    async def execute_many(
        self, sql: str, values: List[Dict[str, Any]], command_timeout: float = None
    ):
        new_sql, args = self._format_sql(sql, **values[0])
        args = [self._format_sql(sql, **v)[1] for v in values]
        return await self._conn.executemany(new_sql, args, timeout=command_timeout)

    def transaction(self, *, isolation=None, readonly=False, deferrable=False):
        return self._conn.transaction(
            isolation=isolation, readonly=readonly, deferrable=deferrable
        )


class Session(AsyncContextManager):
    def __init__(self, pool: Task[Pool], timeout=None) -> None:
        self._pool = pool
        self._timeout = timeout

    async def __aenter__(self):
        pool = await self._pool
        self._conn = _Connection(await pool.acquire(timeout=self._timeout))
        return self._conn

    async def __aexit__(self, extype, ex, tb):
        pool = await self._pool
        await pool.release(self._conn._conn)
        self._conn = None


class Transaction(AsyncContextManager):
    def __init__(
        self, pool: Task[Pool], timeout, isolation, readonly, deferrable
    ) -> None:
        self._pool = pool
        self._timeout = timeout
        self._isolation = isolation
        self._readonly = readonly
        self._deferrable = deferrable

    async def __aenter__(self):
        pool = await self._pool
        conn = self._conn = _Connection(await pool.acquire(timeout=self._timeout))
        tx = self._tx = conn.transaction(
            isolation=self._isolation,
            readonly=self._readonly,
            deferrable=self._deferrable,
        )
        await tx.start()
        return conn

    async def __aexit__(self, extype, ex, tb):
        try:
            if extype == None:
                await self._tx.commit()
            else:
                await self._tx.rollback()
        except:
            pass

        self._tx = None
        pool = await self._pool
        await pool.release(self._conn._conn)
        self._conn = None


class Database:
    def __init__(
        self,
        dsn: str,
        command_timeout=60,
        min_size=10,
        max_size=10,
        max_queries=50000,
        max_inactive_connection_lifetime=300.0,
        loop=None,
    ):
        loop = loop or get_event_loop()

        async def start_pool():
            return await create_pool(
                dsn,
                command_timeout=command_timeout,
                min_size=min_size,
                max_size=max_size,
                max_queries=max_queries,
                max_inactive_connection_lifetime=max_inactive_connection_lifetime,
                loop=loop,
            )

        self._pool = loop.create_task(start_pool())

    async def fetchval(self, sql, col=0, command_timeout: float = None, **kw_args):
        """Return a value based on *sql* query.

        *sql*: sql text, with colon prefixed parameters
        *column*: column of the result set
        *command_timeout*: timeout for sql execution

        Examples:
        >>> db = getfixture('db')
        >>> run_async(db.fetchval("SELECT id, column_1 FROM test", col=1))
        'abc'
        """
        async with Session(self._pool) as session:
            return await session.fetchval(sql, col, command_timeout, **kw_args)

    async def fetch(self, sql, sm: Mapping, command_timeout: float = None, **kw_args):
        """Return multiple rows based on *sql* query.

        *sql*: sql text, with colon prefixed parameters
        *sm*: maps the row into a class
        *command_timeout*: timeout for sql execution

        Examples:
        >>> db = getfixture('db')
        >>> sm = Mapping(dict)
        >>> run_async(db.fetch("SELECT * FROM test where id <= 2 ORDER BY id", sm))
        [{'id': 1, 'column_1': 'abc'}, {'id': 2, 'column_1': 'def'}]
        """
        async with Session(self._pool) as session:
            return await session.fetch(sql, sm, command_timeout, **kw_args)

    async def fetchrow(
        self, sql, sm: Mapping, command_timeout: float = None, **kw_args
    ):
        """Return a row based on *sql* query.

        *sql*: sql text, with colon prefixed parameters
        *sm*: maps the row into a class
        *command_timeout*: timeout for sql execution

        Examples:
        >>> db = getfixture('db')
        >>> sm = Mapping(dict)
        >>> run_async(db.fetchrow("SELECT * FROM test WHERE id=:id", sm, id=2))
        {'id': 2, 'column_1': 'def'}
        """
        async with Session(self._pool) as session:
            return await session.fetchrow(sql, sm, command_timeout, **kw_args)

    async def execute(self, sql: str, command_timeout: float = None, **kw_args):
        """Execute an SQL *command* for each sequence of arguments in *values* list.

        *sql*: sql text, with colon prefixed parameters
        *command_timeout*: timeout for sql execution

        Examples:
        >>> db = getfixture('db')
        >>> run_async(db.execute("INSERT INTO test (id, column_1) VALUES(:id, :col1)", id=103, col1='abcdef'))
        'INSERT 0 1'
        """
        async with Session(self._pool) as session:
            return await session.execute(sql, command_timeout, **kw_args)

    async def execute_many(
        self, sql: str, values: List[Dict[str, Any]], command_timeout: float = None
    ):
        """Execute an SQL *command* for each sequence of arguments in *values* list.

        *values* hold a list of arguments as dict in identical structure. The list has to contain at least one item
        *command_timeout*: timeout for sql execution

        Examples:
        >>> db = getfixture('db')
        >>> values = [dict(id=4, col1='abc'), dict(id=5, col1='def')]
        >>> run_async(db.execute_many("INSERT INTO test (id, column_1) VALUES(:id, :col1)", values))
        """
        async with Session(self._pool) as session:
            return await session.execute_many(sql, values, command_timeout)

    def session(self, timeout=None):
        """Return a database session. It is possible to create transaction from session.

        *timeout*: timeout for getting connection from pool

        Examples:
        >>> db = getfixture('db')
        >>> async def try_transaction():
        ...   async with db.session() as ss:
        ...     async with ss.transaction() as tx:
        ...       await ss.execute("UPDATE test SET column_1=:col", col='useless')
        ...       raise Exception()
        >>> with raises(Exception): run_async(try_transaction())
        """
        return Session(self._pool, timeout)

    def transaction(
        self, *, timeout=None, isolation=None, readonly=False, deferrable=False
    ):
        """Return a database transaction.

        :param timeout: timeout for getting connection from pool

        :param isolation: Transaction isolation mode, can be one of:
                          `'serializable'`, `'repeatable_read'`,
                          `'read_committed'`. If not specified, the behavior
                          is up to the server and session, which is usually
                          ``read_committed``.

        :param readonly: Specifies whether or not this transaction is
                         read-only.

        :param deferrable: Specifies whether or not this transaction is
                           deferrable.
        Examples:
        >>> db = getfixture('db')
        >>> async def try_transaction():
        ...   async with db.transaction() as tx:
        ...     await tx.execute("UPDATE test SET column_1=:col WHERE id=:id", col='useless', id=3)
        >>> run_async(try_transaction())
        >>> run_async(db.fetchval("SELECT column_1 FROM test where id = 3"))
        'useless'
        """
        return Transaction(self._pool, timeout, isolation, readonly, deferrable)

    async def close(self):
        pool = await self._pool
        return await pool.close()



from typing import TypeVar


T = TypeVar("T")


def activerecord(cls: T) -> T:
    """ActiveRecord decorator to keep track of changes happened to a data class

    Examples:
    >>> from dataclasses import dataclass
    >>> @activerecord
    ... @dataclass
    ... class MyRecord:
    ...   col1: int
    ...   col2: str
    >>> record = MyRecord(col1=1, col2="a")
    >>> record.col1 = 2
    >>> record.col1
    2
    >>> record.changes()
    ['col1']
    >>> record.clear()
    >>> record.changes()
    []
    """

    class ActiveRecord(cls):
        def __init__(self, *args, **kargs):
            super().__init__(*args, **kargs)
            self._changeset = set()

        def __setattr__(self, name, value):
            if hasattr(self, "_changeset"):
                self._changeset.add(name)

            return super().__setattr__(name, value)

        def changes(self):
            """Return the properties that have been changed"""
            return [*self._changeset]

        def clear(self):
            """Clear the change log"""
            self._changeset.clear()

    return ActiveRecord



import asyncio
from typing import Coroutine
import uuid
from pytest import fixture, raises
from .dal import SagaDAL, SagaDTO
from .database import Database


@fixture(scope="session")
def db():
    async def run():
        db = Database("postgresql://postgres:1234@localhost:5432/postgres")
        await db.execute(
            "DROP TABLE IF EXISTS test;\
CREATE TABLE IF NOT EXISTS test (id SERIAL PRIMARY KEY, column_1 VARCHAR(100));\
INSERT INTO test (id, column_1) VALUES(1, 'abc');\
INSERT INTO test (id, column_1) VALUES(2, 'def');\
INSERT INTO test (id, column_1) VALUES(3, 'def');"
        )
        return db

    return run_async(run())


@fixture(scope="session")
def dal(db):
    dal = SagaDAL(db)
    run_async(
        db.execute(
            "DROP TABLE IF EXISTS saga.instance;\
DROP TABLE IF EXISTS saga.runner;\
"
        )
    )
    run_async(dal.init_db())
    run_async(
        dal.new_saga(
            SagaDTO(
                id=1,
                name="test",
                instance="random",
                step="step",
                input="abc",
                state="abc",
                runner="runner",
                delay=None,
                status=0,
                error=None,
            )
        )
    )
    run_async(
        dal.new_saga(
            SagaDTO(
                id=2,
                name="event_loop",
                instance="random",
                step="step123",
                input="abc",
                state="abc",
                runner="runner",
                delay=-20,
                status=0,
                error=None,
            )
        )
    )
    return dal


def run_async(co: Coroutine):
    return asyncio.get_event_loop().run_until_complete(co)


@fixture(autouse=True)
def add_np(doctest_namespace):
    doctest_namespace["asyncio"] = asyncio
    doctest_namespace["raises"] = raises
    doctest_namespace["run_async"] = run_async
    doctest_namespace["uuid"] = uuid



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import AsyncGenerator
from .stream_utils import AsyncIterator


class LogLevel(Enum):
    TRACE = "TRACE"
    INFO = "INFO"
    WARN = "WARN"
    ERROR = "ERROR"


@dataclass
class LogEntry:
    # logging data
    time: datetime
    log_level: LogLevel
    message: str
    meta: dict
    error: Exception
    # from logging context
    source: str
    saga_id: int
    saga_name: str
    saga_instance: str
    saga_step: str
    saga_runner: str


class Logger:
    logstream: AsyncIterator[LogEntry]

    def __init__(self, name, runner, logstream=None):
        self.logstream = logstream or AsyncIterator[LogEntry]()
        self._prototype = dict(
            source=None,
            saga_id=None,
            saga_name=name,
            saga_instance=None,
            saga_step=None,
            saga_runner=runner,
        )

    def with_context(
        self,
        saga_id: int,
        saga_instance: str,
        saga_step: str,
        source: str = None,
    ):
        """Return a new logger with context

        Examples:
        >>> logger = Logger('my_logger', 'abc')
        >>> new_logger = logger.with_context(1, 'instance', 'step', 'source')
        >>> new_logger.log_trace('message', data=1)
        >>> log_entry = run_async(logger.logstream.__anext__())
        >>> log_entry.saga_id == 1
        True

        """
        ret = Logger(
            self._prototype["saga_name"], self._prototype["saga_runner"], self.logstream
        )
        ret._prototype["source"] = source or saga_step
        ret._prototype["saga_id"] = saga_id
        ret._prototype["saga_instance"] = saga_instance
        ret._prototype["saga_step"] = saga_step
        return ret

    def _log(self, log_level, message, error, source=None, **kw_args):
        self.logstream.send(
            LogEntry(
                time=datetime.utcnow(),
                log_level=log_level,
                message=message,
                error=error,
                source=source or self._prototype["source"],
                saga_name=self._prototype["saga_name"],
                saga_runner=self._prototype["saga_runner"],
                saga_id=self._prototype["saga_id"],
                saga_instance=self._prototype["saga_instance"],
                saga_step=self._prototype["saga_step"],
                meta=kw_args,
            )
        )

    def log_trace(self, message, error=None, source=None, **kw_args):
        """Log trace level message

        Examples:
        >>> logger = Logger('my_logger', 'abc')
        >>> logger.log_trace('message', error=Exception(), source = 'source', data=1)
        >>> log_entry = run_async(logger.logstream.__anext__())
        >>> log_entry.saga_name == 'my_logger' and log_entry.saga_runner == 'abc'
        True
        >>> log_entry.log_level == LogLevel.TRACE and log_entry.message == 'message' and log_entry.meta["data"] == 1 and repr(log_entry.error) == 'Exception()' and log_entry.source == 'source'
        True
        """
        self._log(LogLevel.TRACE, message, error, source, **kw_args)

    def log_info(self, message, error=None, source=None, **kw_args):
        """Log info level message

        Examples:
        >>> logger = Logger('my_logger', 'abc')
        >>> logger.log_info('message', error=Exception(), data=1)
        >>> log_entry = run_async(logger.logstream.__anext__())
        >>> log_entry.log_level == LogLevel.INFO and log_entry.meta["data"] == 1 and repr(log_entry.error) == 'Exception()'
        True
        """
        self._log(LogLevel.INFO, message, error, source, **kw_args)

    def log_warn(self, message, error=None, source=None, **kw_args):
        """Log warning level message

        Examples:
        >>> logger = Logger('my_logger', 'abc')
        >>> logger.log_warn('message', error=Exception(), data=1)
        >>> log_entry = run_async(logger.logstream.__anext__())
        >>> log_entry.log_level == LogLevel.WARN and log_entry.meta["data"] == 1 and repr(log_entry.error) == 'Exception()'
        True
        """
        self._log(LogLevel.WARN, message, error, source, **kw_args)

    def log_error(self, error, message=None, source=None, **kw_args):
        """Log error level message

        Examples:
        >>> logger = Logger('my_logger', 'abc')
        >>> logger.log_error(error=Exception('message'), data=1)
        >>> log_entry = run_async(logger.logstream.__anext__())
        >>> log_entry.log_level == LogLevel.ERROR and log_entry.meta["data"] == 1 and repr(log_entry.error) == "Exception('message')"
        True
        """
        self._log(LogLevel.ERROR, message or repr(error), error, source, **kw_args)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="pg-saga",
    version="0.1.6",
    scripts=[],
    author="George Ma",
    author_email="z.george.ma@gmail.com",
    description="A minimalist's saga workflow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/z-george-ma/saga",
    packages=["saga"],
    package_dir={"saga": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
